# coding: utf-8

import sys
import os
from glob import glob

try:
    from ConfigParser import SafeConfigParser as ConfigParser, NoOptionError
except ImportError:
    from configparser import ConfigParser, NoOptionError

from .exceptions import InvalidConfigurationFile


class AbstractConfigurationLoader(object):
    def __contains__(self, item):
        raise NotImplementedError()

    def __getitem__(self, item):
        raise NotImplementedError()


class EnvVarConfigurationLoader(AbstractConfigurationLoader):
    def __contains__(self, item):
        return item in os.environ

    def __getitem__(self, item):
        return os.environ[item]


class AbstractFileConfigurationLoader(AbstractConfigurationLoader):
    patterns = ()

    @classmethod
    def get_filenames(cls, path):
        filenames = []
        for pattern in cls.patterns:
            filenames += glob(os.path.join(path, pattern))
        return filenames

    def __getitem__(self, item):
        raise NotImplementedError()

    def __contains__(self, item):
        raise NotImplementedError()


class EnvFileConfigurationLoader(AbstractFileConfigurationLoader):
    patterns = (".env", )

    def __init__(self, filename):
        self.filename = filename
        self.configs = None

    @staticmethod
    def _parse_line(line):
        key = []
        pos = 0
        comment = ""

        # parse key
        for pos, char in enumerate(line):
            if char == "=":
                break

            if char == "#":
                comment = char
                continue

            if comment:
                continue

            if char.isspace():
                continue

            key.append(char)

        else:
            raise ValueError("Invalid line format (key=value)")

        key = "".join(key)

        if not key:
            return

        # parse value
        value = []
        quote = ""
        started = False
        for char in line[pos + 1:]:
            if not char.isspace():
                started = True

            if not started:
                continue

            if char == "#" and not quote:
                break

            if char in "\"'":
                if not quote:
                    quote = char
                    continue

                if quote and quote == char:
                    quote = ""
                    continue

                value.append(char)
                continue

            value.append(char)

        value = "".join(value).rstrip()

        return key, value

    def _parse(self):
        self.configs = {}
        with open(self.filename) as envfile:
            for line in envfile:
                try:
                    key, value = self._parse_line(line.strip())
                except (ValueError, TypeError):
                    continue

                self.configs[key] = value

    def __contains__(self, item):
        if self.configs is None:
            self._parse()

        return item in self.configs

    def __getitem__(self, item):
        if self.configs is None:
            self._parse()

        return self.configs[item]


class IniFileConfigurationLoader(AbstractFileConfigurationLoader):
    default_section = "settings"
    patterns = ("*.ini", "*.cfg")

    def __init__(self, filename, section=None):
        self.filename = filename

        if not section:
            section = self.default_section

        self.section = section

        self.parser = ConfigParser(allow_no_value=True)

        with open(self.filename) as inifile:
            try:
                if sys.version_info[0] < 3:
                    # ConfigParser.readfp is deprecated for Python3, read_file replaces it
                    self.parser.readfp(inifile)
                else:
                    self.parser.read_file(inifile)
            except UnicodeDecodeError:
                raise InvalidConfigurationFile()

        if not self.parser.has_section(self.section):
            raise InvalidConfigurationFile("Missing [{}] section in {}".format(self.section, self.filename))

    def __contains__(self, item):
        return self.parser.has_option(self.section, item)

    def __getitem__(self, item):
        try:
            return self.parser.get(self.section, item)
        except NoOptionError:
            raise KeyError("{!r}".format(item))



# coding: utf-8


from .exceptions import InvalidConfiguration


class AbstractCast(object):
    def __call__(self, value):
        raise NotImplementedError()


class Boolean(AbstractCast):
    default_values = {
        "1": True, "true": True, "yes": True, "y": True, "on": True,
        "0": False, "false": False, "no": False, "n": False, "off": False,
    }

    def __init__(self, values=None):
        self.values = self.default_values
        if isinstance(values, dict):
            self.values.update(values)

    def __call__(self, value):
        try:
            return self.values[str(value).lower()]
        except KeyError:
            raise InvalidConfiguration("Error casting value {!r} to boolean".format(value))


class List(AbstractCast):
    def __init__(self, delimiter=",", quotes="\"'"):
        self.delimiter = delimiter
        self.quotes = quotes

    def _parse(self, string):
        elements = []
        element = []
        quote = ""
        for char in string:
            # open quote
            if char in self.quotes and not quote:
                quote = char
                element.append(char)
                continue

            # close quote
            if char in self.quotes and char == quote:
                quote = ""
                element.append(char)
                continue

            if quote:
                element.append(char)
                continue

            if char == self.delimiter:
                elements.append("".join(element))
                element = []
                continue

            element.append(char)

        # remaining element
        if element:
            elements.append("".join(element))

        return [e.strip() for e in elements]

    def __call__(self, value):
        return self._parse(value)


class Option(AbstractCast):
    """
    Example::
        _INSTALLED_APPS = ("foo", "bar")
        INSTALLED_APPS = config("ENVIRONMENT", default="production", cast=Option({
            "production": _INSTALLED_APPS,
            "local": _INSTALLED_APPS + ("baz",)
        }))
    """
    def __init__(self, options):
        self.options = options

    def __call__(self, value):
        try:
            return self.options[value]
        except KeyError:
            raise InvalidConfiguration("Invalid option {!r}".format(value))



# coding: utf-8


import os
import sys

from .loaders import EnvFileConfigurationLoader, IniFileConfigurationLoader, EnvVarConfigurationLoader
from .exceptions import InvalidConfigurationFile, InvalidPath, InvalidConfigurationCast, UnknownConfiguration
from .casts import Boolean, List, Option


MAGIC_FRAME_DEPTH = 2


class ConfigurationDiscovery(object):
    default_filetypes = (EnvFileConfigurationLoader, IniFileConfigurationLoader)

    def __init__(self, starting_path, filetypes=None, root_path=None):
        """
        Setup the configuration file discovery.

        :param starting_path: The path to begin looking for configuration files
        :param filetypes: tuple with configuration loaders. Defaults to
                          ``(EnvFileConfigurationLoader, IniFileConfigurationLoader)``
        :param root_path: Configuration lookup will stop at the given path. Defaults to
                          the current user directory
        """
        self.starting_path = os.path.realpath(os.path.abspath(starting_path))
        if filetypes is None:
            filetypes = self.default_filetypes
        if root_path is None:
            root_path = os.path.expanduser('~')

        self.root_path = os.path.realpath(root_path)
        self.filetypes = filetypes
        self._config_files = None

        if self.root_path not in self.starting_path:
            raise InvalidPath('Invalid root path given')

    def _scan_path(self, path):
        config_files = []

        for file_type in self.filetypes:
            for filename in file_type.get_filenames(path):
                try:
                    config_files.append(file_type(filename))
                except InvalidConfigurationFile:
                    continue

        return config_files

    def _discover(self):
        self._config_files = []

        path = self.starting_path
        while True:
            if not os.path.isdir(path):
                raise InvalidPath("Invalid path ({})".format(path))

            self._config_files += self._scan_path(path)

            if self._config_files or path == self.root_path or path == os.path.sep:
                break

            path = os.path.dirname(path)

    @property
    def config_files(self):
        if self._config_files is None:
            self._discover()

        return self._config_files


class Configuration(object):
    # Shortcut for standard casts
    boolean = Boolean()
    list = List()
    option = Option

    def __init__(self, configs=None, starting_path=None, root_path=None):
        if configs is None:
            configs = [EnvVarConfigurationLoader()]
        self.configurations = configs
        self.starting_path = starting_path
        self.root_path = root_path

        if starting_path:
            self._init_configs()

    @staticmethod
    def _caller_path():
        # MAGIC! Get the caller's module path.
        # noinspection PyProtectedMember
        frame = sys._getframe(MAGIC_FRAME_DEPTH)
        path = os.path.dirname(frame.f_code.co_filename)
        return path

    def _init_configs(self):
        discovery = ConfigurationDiscovery(self.starting_path, root_path=self.root_path)
        self.configurations.extend(discovery.config_files)

    def __call__(self, item, cast=lambda v: v, **kwargs):
        if not callable(cast):
            raise InvalidConfigurationCast("Cast must be callable")

        if self.starting_path is None:
            self.starting_path = self._caller_path()
            self._init_configs()

        for configuration in self.configurations:
            try:
                return cast(configuration[item])
            except KeyError:
                continue

        if 'default' not in kwargs:
            raise UnknownConfiguration("Configuration '{}' not found".format(item))

        return cast(kwargs["default"])



# coding: utf-8


class ConfigurationException(Exception):
    pass


class InvalidConfigurationFile(ConfigurationException):
    pass


class InvalidPath(ConfigurationException):
    pass


class UnknownConfiguration(ConfigurationException):
    pass


class InvalidConfiguration(ConfigurationException):
    pass


class InvalidConfigurationCast(ConfigurationException):
    pass



# coding: utf-8

import os
import re

from setuptools import setup, Command


here = os.path.abspath(os.path.dirname(__file__))


with open(os.path.join(here, "CHANGES.txt")) as changes:
    for line in changes:
        version = line.strip()
        if re.search('^[0-9]+\.[0-9]+(\.[0-9]+)?$', version):
            break


class VersionCommand(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        print(version)


setup(name='prettyconf',
      version=version,
      description='Separation of settings from code.',
      author="Osvaldo Santana Neto", author_email="prettyconf@osantana.me",
      license="MIT",
      packages=['prettyconf'],
      platforms='any',
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Framework :: Django',
          'Framework :: Flask',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: MIT License',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
          'Topic :: Software Development :: Libraries',
      ],
      url='http://github.com/osantana/prettyconf',
      download_url='https://github.com/osantana/prettyconf/tarball/{}'.format(version),
      cmdclass={'version': VersionCommand},
      test_suite="tests",
)
